/*
 * AD9851-based VFO
 * A PIC16F887 is used to drive an AD9851 DDS via its UART
 *
 * Author: Andrew Fletcher
 * Date:   2017-10-19
 * Location: Edinburgh, Scotland
 *
 */

/*
 * I/O Port usage
 *
 * *** Frequency control
 * RB3 (digital) frequency load word using serial load
 * RD6 --> WCLK
 * RD5 --> FQ_UD
 * RD4 --> RESET
 *
 * Debug
 * RC7 EUSART rx 
 * RC6 EUSART tx
 * RA4 LED
 *
 * Power On LED
 * RA5 
 *
 */
 
//
// Definitions
//
#define VERSION "1.0"

#include <xc.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#define _XTAL_FREQ 8000000

#define NEARLY_ZERO 1

// if these VFO definitions are changed, check the code still works
#define GET_VFO_DELAY 8
#define FREQ_PRE_SCALE 8

#define PON_LED PORTAbits.RA5
#define DEBUG_LED PORTAbits.RA4

#define MAX_LINE_LENGTH 64

//
// DDS definitions
//
#define DDS_W_CLK PORTDbits.RD6
#define DDS_FQ_UD PORTDbits.RD5
#define DDS_DATA  PORTBbits.RB3
#define DDS_RESET PORTDbits.RD4
#define DDS_CLOCK 30000000
#define DDS_REFMUL 0x01
#define MAX_HERTZ 52000000
#define DEFAULT_HERTZ 2000000
// pre-calculated pow(2, 24) saves code space and increases speed
#define POW_2_24 16777216
#define MAX_FREQ_DELTA 4

//
// Configuration bits (PIC16F887) 
// TODO properly
#pragma config WDTE = OFF, PWRTE = OFF, CP = OFF, BOREN = ON, FCMEN = ON, MCLRE = OFF, CPD = OFF, IESO = ON, FOSC = INTRC_CLKOUT, LVP = OFF

//
// global variables
//
// None

//
// ISRs
//
// None

//
// production functions
//

void setClock(void) {
  // set PIC internal oscillator to 8MHz
  OSCCONbits.IRCF0 = OSCCONbits.IRCF1 = OSCCONbits.IRCF2 = 1;

  // and tune to around 2MHz
  OSCTUNEbits.TUN4 = 0;
  OSCTUNEbits.TUN3 = 0;
  OSCTUNEbits.TUN2 = 0;
  OSCTUNEbits.TUN1 = 0;
  OSCTUNEbits.TUN0 = 0; 
}

void loadFTW(unsigned long data) 
{
unsigned long pointer = 0x01;

    DDS_FQ_UD = 0;

	// load frequency
	for (int i=0; i<32; i++) {
      DDS_DATA = ((data & pointer) > 0); 
		__delay_ms(1);
		DDS_W_CLK = 1;
		__delay_ms(1);
		DDS_W_CLK = 0;
		pointer = pointer << 1;
		__delay_ms(1);
	}
 
	pointer = 0b00000001;
 
 	// load reference multiplier and phase
	for (int i=0; i<8; i++) {
      DDS_DATA = ((DDS_REFMUL & pointer) > 0);
        __delay_ms(1);
        DDS_W_CLK = 1;
    	__delay_ms(1);
    	DDS_W_CLK = 0;
    	pointer = pointer << 1 ;
   		__delay_ms(1);
  	}
  	
	__delay_ms(1);
	DDS_FQ_UD = 1;
	__delay_ms(1);
	DDS_FQ_UD = 0;
}

 
void setupDDS() 
{  
  // initialise DDS control signals
  DDS_RESET = DDS_W_CLK = DDS_FQ_UD = DDS_DATA = 0;
  __delay_ms(1);
  
  // reset the DDS
  DDS_RESET = 1;
  __delay_ms(1);
  DDS_RESET = 0;
  __delay_ms(1);
  
  // enable serial mode (assumes D2 is pulled down, D0 and D1 are pulled up)
  DDS_W_CLK = 1;
  __delay_ms(1);
  DDS_W_CLK = 0;
  __delay_ms(1);
  
  DDS_FQ_UD = 1;
  __delay_ms(1);
  DDS_FQ_UD = 0;
  __delay_ms(5);
}

void enableEUSARTtx(void) {
  // set EUSART baud rate to 19200, assuming a 8MHz clock
  TXSTAbits.BRGH = 1;
  BRG16 = 0;
  SPBRGH = 0;
  SPBRG = 25;  
  
  // enable EUSART TX
  TXSTAbits.TXEN = 1;
  TXSTAbits.SYNC = 0;
  RCSTAbits.SPEN = 1;
}

// put a character on the EUSART 
void putcharEUSART(unsigned char ch) {
	while(!PIR1bits.TXIF); 
    TXREG = ch;
} 

// put a string on on the EUSART
void printEUSART(const unsigned char str[]) {
unsigned i = 0;
	while (str[i]) {
		putcharEUSART(str[i++]);
	}
}

void enableEUSARTrx(void) {
    RCSTAbits.RX9 = 0;    // 8 bit mode
    RCSTAbits.CREN = 1;   // Enable receive
    RCSTAbits.ADDEN = 0;  // Disable address detection
    RCIE = 0;             // no interrupts
}

char EUSARTdataReady() {
  return RCIF;
}

char getcharEUSART() {
  while(!RCIF);
  return RCREG;
}

void getlineEUSART(char *line, unsigned int maxlength) {
  for (int i=0; i < maxlength; i++) {
    line[i] = getcharEUSART();	
    putcharEUSART(line[i]); // echo
    if ((line[i] == '\r') || (line[i] == '\n')) {
    	line[i] = '\0';
    	return;
    }
  }
  line[maxlength - 1] = '\0';
}

//
// main
//
void main(void) {
double deltaPhase;
unsigned char longStr[11];
unsigned long frequency;
char line[MAX_LINE_LENGTH];
char * dummy;

  // disable all interrupts
  INTCON = 0;
  
  // configure i/o ports
  ANSEL = ANSELH = 0;            // configure all as digital unless needed otherwise
  
  CM1CON0 = CM2CON0 = 0;         // disable comparators
  
  TRISA = 0xFF;		    // all of PORTA are inputs
  TRISAbits.TRISA5 = 0; // except RA5 (PON LED)
  TRISAbits.TRISA4 = 0; // and RA4 (Debug LED)
  PON_LED = 1;
  
  TRISB = 0xFF;            // all of PORTB are inputs
  TRISBbits.TRISB3 = 0;    // except RB3 used for DDS serial load
  
  TRISC = 0xFF;            // all of PORTC are inputs
  TRISCbits.TRISC6 = 0;    // except EUSART TX
  
  TRISD = 0xFF;            // all of PORTD are inputs
  TRISDbits.TRISD4 = TRISDbits.TRISD5 = TRISDbits.TRISD6 = 0;
                           // except RD4-6 used for DDS

  TRISE = 0xFF;  		// all of PORTE are inputs
  
  setClock();           // set clock to 8MHz
  
  enableEUSARTtx();
  enableEUSARTrx();
  
  setupDDS();           // put DDS in serial mode and load null FTW
  loadFTW(0);
  
  printEUSART("\r\n\r\nAD9851 VFO ");
  printEUSART(VERSION);
  printEUSART(" by Andrew Fletcher\r\n");
  
  while (1) {
  	  printEUSART("Frequency: ");
  	  getlineEUSART(line, MAX_LINE_LENGTH); // note: blocks execution
  	  frequency = strtol(line, &dummy, 10);

  	  DEBUG_LED = 0;

      // derive delta phase in a way to minimise the code space
      // Fout = (delta Phase x System Clock) / 2^32
      deltaPhase = POW_2_24;
      deltaPhase *= frequency;
      deltaPhase /= (DDS_CLOCK / 256);
      deltaPhase /= (DDS_REFMUL & 1 ? 6 : 1);
      loadFTW((unsigned long)deltaPhase);
      
      DEBUG_LED = 1;
      
      printEUSART("\r\nloaded frequency of ");
      ultoa(longStr, frequency, 10); 
	  printEUSART(longStr);
	  printEUSART(", deltaPhase = ");
	  ultoa(longStr, (unsigned long)deltaPhase, 16); 
	  printEUSART(longStr);
	  printEUSART("H\r\n");
    }
}
